# FisbCusumFuncs.r
#  UE project FISBOAT n 502572
#  B. Mesnil & P. Petitgas, Ifremer, France
#  Last Update: 19/01/2007
#
#  Suite of functions for using cusum charts in the FISBOAT project
#
#  This file should be sourced into your (case study) workspace
#
#  Three blocks of functions:
#	1. Computing decision-interval or simple Cusums
#	2. Tuning a Cusum chart : give ARL & its distribution
#	3. Auxiliaries & second choice
#
# =================
#  Functions for computing the Cusum
# =================
dicusum.f <- function(data,cck,cmu,csd) {
# Computes cusum phi+ and phi- as in Scandol 2003,2005
#  i.e. in "decision interval form" or "tabular form" (Hawkins & Olwell 1997)
#  data = Raw-data vector (standardisation done inside this function),
#  cck = allowance k, in sd units ; cmu & csd = in-control mean & sd

nd <- length(data)+1
phip <- vector(mode="numeric",length=nd)
phim <- vector(mode="numeric",length=nd)

# first standardise data
z <- (data - cmu)/csd
# recursion
phip[1] <- 0
phim[1] <- 0
        for(i in 2:nd) {
	            phip[i] <- max(0,(phip[i-1]+z[i-1]-cck),na.rm=TRUE)
	            phim[i] <- min(0,(phim[i-1]+z[i-1]+cck),na.rm=TRUE)
             }
cbind(phip[2:nd],phim[2:nd])

}

#
# same dicusum function but with plots of the charts inside the function
#
cusum.decide<-function(tt,yy,m.ref,s.ref,k.allow,h.decide) {
# after: Scandol, J. 2003. Fisheries Research, 64: 19-36.
# tt, yy: the data time series where tt is time and yy the variable
# m.ref, s.ref: mean and std.dev of xx estimated in a reference period
# k.allow: allowance +/- k (non important deviations)
# h.decide: decision interval +/- h (signal of out-of-control to be triggered)
# if (phi.plus > h) or (phi.minus < -h): trigger alert signal
 nn<-length(yy)
 zz<-(yy-m.ref)/s.ref
 phi.plus<-rep(0,(nn+1)); phi.minus<-rep(0,(nn+1))
 for (i in 2:(nn+1)) { phi.plus[i]<-max(0,(phi.plus[i-1]+zz[i-1]-k.allow),na.rm=T)
                       phi.minus[i]<-min(0,(phi.minus[i-1]+zz[i-1]+k.allow),na.rm=T) }
 phi.plus<-phi.plus[-1]; phi.minus<-phi.minus[-1]
 x11(); plot(tt,phi.plus,xlab='',ylab='di-cusum'),
             type="b",
             ylim=c(-max(c(abs(phi.minus),phi.plus)),max(c(abs(phi.minus),phi.plus))))
        points(tt,phi.minus,type="b")
        abline(0,0,lty=2);abline(h.decide,0,lty=2);abline(-h.decide,0,lty=2)
 return(cbind(tt,yy,phi.plus,phi.minus))
}


#
# =================
#  Functions for tuning the Cusum
# =================
#  The functions are transcript of a Fortran program by F.F. Gan (1993)
#   J. Quality Technology 25(3):205-215 found on Statlib /JQT archive
#  !! no serious attempt to vectorise & get rid of for() loops !!
#  !! dirty code to be tidied up later (suggestions welcome)   !!
# ===========
cusarl.f <- function(cck,cch,shift,hs) {
#   Computes ARL of 1-sided Cusum for shift 'shift', head-start hs 
#    reference cck and decision limit h
#   Using algorithm by F.F. Gan (1993), JQT 25(3):205-215
#    found on Statlib /JQT archive
#
np <- 25
A <- matrix(nrow=np,ncol=np)
nq <- np-1
#  Get Gauss quadrature
gq <- gq.f(cch)

#   setup A matrix of linear system 
for(i in 1:np){
  for(j in 1:np) {
	if(i <= nq && j <= nq) {
	 arg <- gq[j,1]+cck-gq[i,1]
	 A[i,j] <- gq[j,2]*dnorm(arg,shift,1) }
	if(i <= nq && j==i) {  # Moved here, otherwise diagonal overwritten
       arg <- gq[j,1]+cck-gq[i,1]
	 A[i,j] <- gq[j,2]*dnorm(arg,shift,1) -1 }
	if(i <= nq && j== np) A[i,j] <- pnorm(cck-gq[i,1]-shift)
	if(i == np && j < i) {
	 arg <- gq[j,1]+cck
	 A[i,j] <- gq[j,2]*dnorm(arg,shift,1) }
	if(i == np && j == i) A[i,j] <- pnorm(cck-shift)-1
  }
}
##return(A) # checked OK so far
#  Solve linear system
B <- rep(-1,25)
x <- solve(A,B)
# ARL
arl <- 1+ x[np]*pnorm(cck-hs-shift)
	for(j in 1:nq) {
	arg <- gq[j,1]+cck-hs
	arl <- arl + gq[j,2]*dnorm(arg,shift,1)*x[j]
	}
return(round(arl,1))
}

# ===============
arldis.f <- function(cck,cch,shift,hs) {
# computes Run Length distribution of 1-sided Cusum 
#  calls cusp.f, gq.f & squant.f
nmx <- 999
cdf <- vector("numeric",length=nmx)
run <- vector("integer",length=nmx)
prob <- seq(1:nmx)/1000
probs <- c(0.01,0.05,0.25,0.5,0.75,0.95,0.99)
qtl <- rep(NA,length(probs))

cat(" Computing Run Length Distribution - be patient!\n")
flush.console()
gq <- gq.f(cch) # call Gauss quadrature
pr <- vector("numeric",length=dim(gq)[1])
cum <- 0
kount <- 1
for(n in 1:nmx) {
      x <- cusp.f(n,cck,cch,shift,hs,gq,pr)
      cum <- cum+x$ppp
      run[n] <- n
      prob[n] <- x$ppp
      cdf[n] <- cum
   		if(cum > 0.999) break
	# extract pr to re-pass it to cusp.f
	pr <- x$pr
	} # end for loop
# Display key percentiles 
for(i in 1:length(probs[probs<=cum])) {qtl[i] <- squant.f(cdf[1:n],probs[i]) }
cat("Key quantiles of the RL distribution\n",
"    1% 5% 25% 50% 75% 95% 99% \n")
print(qtl)  # failed attempts at pretty formatting

# Plots
sout <- paste("k =",cck,"; h =",cch,"; shift =",shift,"; head =",hs,sep=" ")
windows()
plot(run[1:n],cdf[1:n],type="l",main=paste("CDF of 1-sided Cusum RL\n",sout),
xlab="Run Length",ylab="Cumul. Probability")
windows()
plot(run[1:n],prob[1:n],type="l",main=paste("pdf of 1-sided Cusum RL\n",sout),
xlab="Run Length",ylab="Probability",las=1)
abline(v=cusarl.f(cck,cch,shift,hs),lty=2)
  
# return results in dataframe, e.g. to redo plotting
data.frame(Run=run[1:n],Prob=prob[1:n],CDF=cdf[1:n])
}

# =================
gq.f <- function(h) {
#  maps Gauss quadrature points from (-1,1) to (0,h)
#   assumes 24 points
#  Gan's SETUP routine
#
NQUAD <- 24
PT <- vector("numeric",length=NQUAD+1)
WT <- vector("numeric",length=NQUAD+1)
P <- c(0.9951872199970213,0.9747285559713095,0.9382745520027327,
0.8864155270044010,0.8200019859739029,0.7401241915785543,0.6480936519369755,
0.5454214713888395,0.4337935076260451,0.3150426796961634,0.1911188674736163,
0.0640568928626056,0,0,0,0,0,0,0,0,0,0,0,0,0)
W <- c(0.0123412297999872,0.0285313886289337,0.0442774388174198,
0.0592985849154368,0.0733464814110803,0.0861901615319533,0.0976186521041139,
0.1074442701159656,0.1155056680537256,0.1216704729278034,0.1258374563468283,
0.1279381953467521,0,0,0,0,0,0,0,0,0,0,0,0,0)
ND2 <- NQUAD/2
NQP1 <- NQUAD+1

	for(i in 1:ND2) {
	P[NQP1-i] <- -P[i]
	W[NQP1-i] <- W[i]
	}
WT <- h*W/2
PT <- h*P/2 + h/2
cbind(PT,WT)
}

# ===============
cusp.f <- function(n,cck,cch,shift,hs,gq,pr) {
#  routine used to obtain ARL distribution ; called by arldis.f()
#  Computes the probability function of the run length
#   from F.F. Gan
#  pr needs to be passed back & forth (updated in each pass)

np <- 25
nq <- np-1
prx <- vector("numeric",length=np)

if(n == 1) {
arg <- cch+cck-gq[,1]-shift # np terms  !!
pr <- 1- pnorm(arg)
pr[np] <- 1- pnorm(cch+cck-shift)  # change last one
ppp <- 1- pnorm(cch+cck-hs-shift)
#print(pr)
}  else {
# now case n >=2
prx <- pr
for(j in 1:nq){
pr[j] <- prx[np]*pnorm(cck-gq[j,1]-shift)
      for(i in 1:nq) {
      pr[j] <- pr[j]+ gq[i,2]*prx[i]*dnorm(gq[i,1]+cck-gq[j,1],shift,1)
      }
}
pr[np] <- prx[np]*pnorm(cck-shift)
for(i in 1:nq) {
pr[np] <- pr[np]+ gq[i,2]*prx[i]*dnorm(gq[i,1]+cck,shift,1)
#
ppp <- prx[np]*pnorm(cck-hs-shift)
    for(i in 1:nq) {
    ppp <- ppp+ gq[i,2]*prx[i]*dnorm(gq[i,1]+cck-hs)
    }
}
}
if(ppp < 1.0e-20) ppp <- 0
out <- list(ppp=ppp,pr=pr)
return (out)
}

# ===============
squant.f <- function(vec,prob){
# get rank of value prob in vector vec of cumulated probs, does not interpolate
# !! terrible code !!
i <- 1
while(vec[i]<prob) {i <- i+1}
return(i)
}

# ===============
#  Auxiliaries
# ===============
#  !! this is a second choice - prefer dicusum.f above !!
cusum.f <- function(data,cmu) {
# Computes "traditional" cusum 
#   note: data not standardised
nd <- length(data)
S <- vector(mode="numeric",length=nd)
# recursion
S[1] <- data[1]-cmu
for(i in 2:nd) {
	S[i] <- S[i-1] + (data[i]-cmu)
	}
return(S)
}

# =========
SiegARL.f <- function(k,h,shift,type) {
# Function to calculate the ARL of a (k,h) 1-sided Cusum
#  using Siegmund's approximation
#  k, h and shift are in sd units
#  type= 1 for Upper, = 2 for Lower Cusum
b <- h+1.166
if(type==1) delta <- shift-k else delta <- -shift-k
if(delta == 0) A <- b*b else {
x <- 2*delta*b
A <- (exp(-x)+x-1)/(2*delta*delta)
}
A
}

#=======
shewart.xbar<-function(tt,yy,m.ref,s.ref,lim) {
 u<-(yy-m.ref)/s.ref
 x11();plot(tt,u,type="b",xlab="",ylab="(x-m)/sd") 
 abline(h=-lim,lty=2);abline(h=lim,lty=2)
}

#========
# functions used in Fisboat applications 
#========

which.deviation<-function(yy,sely) {
# returns the average deviation from mean m in sd units outside the ref period
# m & sd are for the ref period
 mm<-mean(yy[sely],na.rm=T); ss<-sd(yy[sely],na.rm=T)
 dev<-mean(abs(yy[!sely]-mm)/ss,na.rm=T)
 cat("\n","in ref period: m : ",mm,"    sd : ",ss,"\n",
    "out of ref period : average dev from m in sd units : ",dev,"\n")
 return(dev)
}

tune.cusum<-function(cckk,cchh) {
# prints on the screen the key parameters for tuning the cusum to detect a given deviation
# deviation is that from m in sd units, where m and sd are for the ref period
 ic<-cusarl.f(cckk,cchh,0,0); cat("in-control ARL: ",round(ic,1),"\n") 		
 ix.rl<-arldis.f(cckk,cchh,0,0); ic25<-ix.rl$Run[order(abs(0.25-ix.rl$CDF))[1]]
 oc<-cusarl.f(cckk,cchh,2*cckk,0); cat("out-of-control ARL: ",round(oc,1),"\n")
 return(c("k"=cckk,"h"=cchh,"icarl"=ic,"icrl25"=ic25,"shift"=2*cckk,"ocarl"=oc)) 
}





