# FisbCusum.r
#   UE project FISBOAT n502572
#   B. Mesnil & P. Petitgas, Ifremer, France
#   Last Update: 17/11/2006 - 28/08/2007 (simulations)
#
#   A script implementing CuSum schemes to detect changes in mean level
#    in a time series of continuous (as opposed to discrete) indicators
#     for exploration in FISBOAT project
#   Divided in 2 parts (+ installation):
#	I. Computing & plotting a Cusum, given the parameters
#	II. Tuning the parameters for desired performance
#
# NOTE: this script is for use in "interactive mode":
#       highlight & submit code step by step,
#       after editing with specifics of your case
#  Some optional commands are commented out to prevent damage: just send the bits without the initial #
stop("Only to be run stepwise")  # DO NOT RUN IN ONE GO

#   INSTALLATION
# =================
#  Check the working directory, where your (case study) data should be stored
getwd()
#  Set the correct directory if needed
#setwd("C:/blablabla")  # adjust address

# On first use, load the code of functions in file CusumFuncs.r into your workspace;
#  if this file has already been copied in your working directory, just do:
source("CusumFuncs.r")
otherwise:
#source("C:/where-it-is/CusumFuncs.r")  #  adjust path

## If you only want to play with tuning, the indicators data are not needed
##  and you can skip Part I.

#   I. GETTING A CUSUM PLOT
# ==========================
# It is assumed that the indicator series you want to cusum-check
#   is accessible in your workspace, otherwise you need to load it
#  (bits borrowed from Verena Trenkel's scripts - Adjust path & species' name)
#indicest<-read.table("DataTables/codNS_tab2_wp2a.txt",header=T,sep="\t",as.is=T)
#speciesname<-unique(indicest$Species)
#id<-indicest$Species==speciesname
#Years<-indicest$Year[id]

# First select the particular indicator into working object myind, e.g.
#names(indicest)    # if you have forgotten
myind <- indicest$Recruit
#  sometimes useful to scale down, for prettier labels on plots, e.g.
#myind <- myind/1000  # !! adjust for your case
# and specify years of full series (unless done above)
Years <- 1985:2005       # or 1:21 or ... as you like

# Remember: CuSum assumes Normality => You may need to transform your data, e.g.
# myind <- log(myind)   # or any other adequate transformation

# The crucial next step is to define a reference period over which
#  the "in control" mean & sd are computed
#  in this example, first 10 years are taken as reference
Refy <- 1985:1994    # or Years[1:10]  or c(1985, ....) # Adjust
irefy<-is.element(unique(indicest$Year),Refy)
cmu <- mean(myind[irefy],na.rm=TRUE)
csd <- sd(myind[irefy],na.rm=TRUE)

#  Then choose the parameter k (= allowance, reference value), in sd units;
#   the rationale is to set k = Half the shift that you want to detect.
#  If you have no idea for the size of expected shifts, look at the mean standardised
#   shifts from cmu OUTside the reference period, INDICATIVE ONLY! 
abs(mean((myind[!irefy]-cmu)/csd,na.rm=TRUE))

cck <- 0.5   # example ; adjust

# In principle, one should also set a threshold H (cch) which, when exceeded,
#  triggers an alarm - this is dependent upon managers' perception of
#  acceptable vs. unacceptable risk. At this stage, we leave the question
#  open and H is not set nor indicated on the graphs ; easy to add an
#   abline(h=Hvalue) in big red whenever consensual H exists.

# then call function - returns a 2-col array phi+,phi-
phi <- dicusum.f(myind,cck,cmu,csd)

# give a title and submit plot commands
titre <- "NS cod - Log(Recr. index)"   # adjust
par(mar=c(3,5,3,5)+0.1)
matplot(Years,phi,type="l",lty=1,col=1,main=titre,xlab="Year",
ylab=expression(paste(phi,"-/",phi,"+")),tck=-0.01,las=1)

# optional, show decision limits +H|-H  !! IF they fit in the plot !!
#cch <- 3   # for example
#abline(h= cch,lty=2,col=2)
#abline(h= -cch,lty=2,col=2)

# If you wish, superpose raw data ; send the block below
par(new=TRUE)
plot(Years,myind,axes=FALSE,type="p",pch=20,xlab="",ylab="",col="blue")
axis(side=4,las=1,tck=-0.01)
mtext(side=4,line=3.8,"Raw index")
par(new=FALSE)


#   II. TUNING A CUSUM SCHEME
# ===========================
#  "Tuning" means selecting the Cusum parameters k & h for desired performance
# The rationale for choosing k is based on the magnitude of the shift in mean that
#  makes "a meaningful impact" on the system, driving it out of control -
#  then you set k = that.shift /2;
# The rationale for h is based on the average run length, i.e. the average
#  number of observations between the start of monitoring and the occurrence 
#  of an alarm - an alarm occurs when the cusum plot crosses the horizontal line at h
# It is suggested to proceed in 3 steps:
#	1. Set shift to 0, and try an h that gives a large ARL to a false alarm (e.g. 25 or more)
#	2. Look at the run length distribution, and check that the probability of
#	   short run lengths to false alarms is small enough (a 25% percentile < 10 is annoying)
#	3. Once happy with these settings, set shift = 2*k and check that the ARL for 
#	   such an out-of-control case is as small as possible (e.g. 2-4)
#  re-iterate until the Cusum design performs as desired

#  Step 1
# --------
#  Set the values of k, h, shift & head-start  (all as multiples of s.d.)
cck <- 0.5   # allowance k
cch <- 1.5   # decision limit h, industry folks often consider 4 or 5
shift <- 0   # = 0 for In-Control ARL in Step 1
hs <- 0      # head-start, normally leave it = 0
#  get the in-control ARL(0) with function cusarl.f
cusarl.f(cck,cch,shift,hs)
# alternatively, you can play with values directly in the call
cusarl.f(0.5,1.5,0,0)

#  Step 2
# --------
#  Check the quantiles & distribution of ARL(0)
#  Call to arldis.f returns a table of key quantiles on the R Console and 2 graphs:
#   one of R.L. probability distribution + a vertical dashed line at the ARL;
#   and one of the cumulative probability (hidden behind the former);
#    the distributions are calculated (truncated) up to a run length of 999
#  a data frame of results is returned in x
#   !! be patient: it takes some time !!
x <- arldis.f(cck,cch,shift,hs)
#   take a larger h if lower quartile & median in the table are too small

#  Step 3
# --------
#  Check that the out-of-control ARL is small enough
#   a shift = 2*cck is assumed (3rd argument), you can change to another value
cusarl.f(cck,cch,2*cck,hs)


# ===============
#   EXTRAS = Optional
#  ========
#  Traditional CuSum, as a second choice
#   only consider deviations from desired mu; here equated with cmu set above
#   apparently, data used as they are (not standardised)

# call function cusum.f on your data (specified in Part I above)
su <- cusum.f(myind,cmu)

# give title and plot
titre <- "NS cod - Log(Recr. index)"
plot(Years,su,type="b",xlab="Year",ylab="CuSum",
las=1,main=title)
abline(h=0,lty=2)
# if you wish, draw a smoother through
lines(smooth.spline(Years,su),lty=3)
#lines(lowess(su),lty=2,col="blue")  # alternative choice for smoother

# ==============
#  Siegmund approximation for ARL of 1-sided Cusum
# recall k, h and shift are in sd units
# type= 1 for Upper, = 2 for Lower Cusum
# args. are (k,h,shift,type)
SiegARL.f(cck,cch,shift,1)

# ==============
#  Code to redo the plots made by arldis.f() from the returned object x,
#   if necessary, without running the function again
sout <- paste("k =",cck,"; h =",cch,"; shift =",shift,"; head =",hs,sep=" ")
#tmp <- "k = 0.5 ; h = 3 ; shift = 0 ; head = 0"
plot(x$Run,x$Prob,type="l",main=paste("pdf of 1-sided Cusum RL\n",sout),
xlab="Run Length",ylab="Probability",cex.lab=1.2,) #las=1
abline(v=cusarl.f(cck,cch,shift,hs),lty=2)
# this second plot is perhaps of less interest
windows()
plot(x$Run,x$CDF,type="l",main=paste("CDF of 1-sided Cusum ARL\n",sout),
xlab="Run Length",ylab="Cumul. Probability")

# ==============
# Try a Shewhart Control Chart
par(mar=c(3,6,4,2)+0.1)
plot(Years,(myind-cmu)/csd,type="b",xlab="Years",ylab=expression(frac(x-mu,sigma)),
main=paste(titre,"\nShewhart Chart"))
# and show the 3-sigma control limits, if they fit !!
abline(h=3,lty=2)
abline(h=-3,lty=2)


#  End - clean up workspace (if you wish)
# ==============
#rm(myind,Years,Refy,irefy,cmu,csd,cck,cch,shift,hs,phi,titre)
#rm(x,su,sout)
# if you need to remove the functions, all are named with a suffix .f
ls(pattern="\.f")   # WATCH: even with \., returns all names containing f
grep(".f",ls(),value=T,fixed=T)  # works


# ======= Aug 07 =======
# Check simulation procedure for ARL of UPPER Cusum of Normal data
#  You need to load the function below
# call:  Watch! Does not handle headstart
sim.arl.norm.f(0.5,1.5,0.0)
x <- sim.arl.norm.f(0.5,1.5,2*cck)
# compare
cusarl.f(0.5,1.5,shift,0)
x <- arldis.f(cck,cch,shift,hs)

# --- function ---
sim.arl.norm.f<-function(cck,cch,shift) {
#  ARL of Upper Sided Cusum for Normal data
# estimation by simulation using a variance reduction technique with total hazards
# Jun and Choi. 1993. Commun. Statist. Simula. 22: 877-887
# mean(rl) using nr=5000 realisations of rl
# var(arl)=mean[(rl-arl)2] estimated using 100 resampled rl values, with 100 repetitions
#  Adapted from Pierre's code for chi-2 (summer 2007) 
 
nr<-5000     # simulation runs
nn<-rep(0,nr);yy<-rep(0,nr)
nboot<-100   # boot size for variance
d1boot<-rep(0,nboot)

for (k in 1:nr) {  # simulate N's and control variates Y
	s<-0;i<-0;p<-0
 	while(s<cch) {i<-i+1
              p<-p+1-pnorm(cck-shift+cch-s) 
              x<-rnorm(1) 
              s<-max(0,s+x-cck+shift) }
 	nn[k]<-i; yy[k]<-p }
dd1<-mean(nn)-cov(nn,yy)*(mean(yy)-1)/var(yy)
qq25<-quantile(nn,0.25);  qq75<-quantile(nn,0.75)
# variance
for (k in 1:nboot) {
	ii<-sample(1:nr, size=100, replace=T, prob = NULL)
	d1boot[k]<-mean(nn[ii])-cov(nn[ii],yy[ii])*(mean(yy[ii])-1)/var(yy[ii]) }
ssd1<-sqrt(var(d1boot))
# graph
sout <- paste("k =",cck,"; h =",cch,"; shift =",shift,sep=" ")
x11()
hist(nn,freq=F,nclas=100,main=paste("pdf of 1-sided Cusum RL\n",sout),xlab='ARL')
abline(v=dd1,lty=2)
cat("ARL: ",round(dd1,2), "; sd: ", round(ssd1,2),"; q25: ",round(qq25,2) ,"; q75: ",round(qq75,2) ,"\n")
res<-c(ic.arl=dd1,sd=ssd1,q25=qq25,q75=qq75)
 return(res)
}





