# FBCusumCharts.R : decision interval cusum charts for Fisboat indicators
# UE project FISBOAT n502572
# P. Petitgas, B. Mesnil, Ifremer
# 17/11/2006
#
# 
##########
# Guide lines for tuning the cusum chart
# chose parameters cck and cch to obtain long (>20y) IC ARL 
#                                         long (>10y) IC 25 percentil RL (RL distribution is skewed)
#                                         small (<3y) OC ARL for shift=2*cck 
# by undergoing iteratively steps 1 to 3 with function tune.cusum
# 1. set value of deviation from m underwhich no attention is paid
# input cck (cck=deviation/2)
# defaults are : cck = 0.5 mean deviation from m in sd units (m & sd for ref period)
#                cck = sd
# 2. tune false alarm rate
# compute in-control ARL(0) (IC ARL) and the distribution of the IC RL
# set cch (no default) so that IC RL is large enough (ARL & 25 percentile RL)
# shift & headstart are defaulted : shift=0, hs=0
# no default for cch 
# 3. check detection capability (power) once shift has occurred 
# compute out-of-control ARL (OC ARL) for a given shift from m
# OC ARL should be small, e.g. <3y. if not, change cch or cck
# defaults are: shift=2*cck & hs=0
##########
# Main :
# Source the cusum functions
# Read the Fisboat tables containing indicators time series
# Plots the indicator time series
# Define reference period 
# Tune the cusum
# Plot the DI-cusum
# Save tuning parameters and cusum alarms 
# inputs: 
#         pat for sourcing FisbCusumFuncs.r
#         pat, kfic for reading the file
#         jcol for locating indicator column in the file
#         resnam for saving results
#         cck, cch for tuning the di-cusum 
# outputs:
#         on screen: key parameters for tuning
#         graphs: indicator time series, 
#                 cusum ARL distribution (pdf & cdf)
#                 decision interval cusum (dicusum)
#         object for saving tuning parameters and alarm signal
#
# the code is written for wp2a tab2 Fisboat tables for vital traits
# modify appropriatly parameters pat, kfic, jcol & resnam when working on 
# wp2a tab1 Fisboat tables for spatial indicators 
########## 

# source functions CusumFuncs.r
#
# input appropriate path to specify where CusumFuncs.r is located
pat<-"C:/ptg/WORK/fp6_fisboat/Methods_WP5indicators/cusum_BenoitPierre/"
source(paste(pat,"CusumFuncs.r",sep=""))

# read data

# names of Fisboat Indicator Tables containing the indicators time series 
fic<-c("ancBB_ac_tab2_wp2a.txt",
       "mulTS10ab_tab2_wp2a.txt",
       "herNS_tab2_wp2a.txt",
       "codNS_tab2_wp2a.txt","codBS_tab2_wp2a.txt","codBA_tab2_wp2a.txt",
       "hakBB_tab2_wp2a.txt","hakIO_tab2_wp2a.txt","hakAE_tab2_wp2a.txt")

# input appropriate path pat and index k for case study in fic
pat<-"C:/ptg/WORK/fp6_fisboat/wp2a_tables/wp2a_tab2/"
kfic<-1

# read data  
tab<-read.table(paste(pat,fic[kfic],sep=""),header=T,sep='\t',as.is=T)
nl<-length(tab[,1]); nc<-length(tab[1,])
cat("\n","first line of file ","\n"); tab[1,]
colnam<-dimnames(tab)[[2]]

# input name of object for saving results # change if necessary 
resnam<-"ancBBac.cusumtab2"

# plot indicator time series

# input appropriate column j of indicator table corresponding to your indicator of interest
jcol<-5 
if (jcol ==5 | jcol==6) { y<-log(tab[,jcol])
 x11(); plot(tab$Year,y,type="b",pch=18,ylab=paste("Ln-",colnam[jcol],sep=""),
             xlab="") }
if (jcol>6) { y<-tab[,jcol]
 x11(); plot(tab$Year,y,type="b",pch=18,ylab=colnam[jcol],xlab="") }

# define reference period: input Ref

# reference periods as defined at the Fisboat IJmuiden meeting
Refy<-list(1990:2001,
           1999:2003,
           1989:1993,
           1985:1994,1996:2004,c(1994:1996,1998:1999),
           1987:1996,1998:2001,1998:2001)

# if ref period in Refy not appropriate for indicator, input instead a vector of years in Ref
Ref<-Refy[[kfic]]         
sel.refy<-is.element(unique(tab$Year),Ref)

# tune the di-cusum

dev<-which.deviation(y,sel.refy); cat("proposed cck : ",round(dev/2,1),"\n")

# input iteratively cck and cch until satisfaction (follow guide lines above)
cck<-1	# change as necessary
cch<-1	# change cch as necessary
param<-tune.cusum(cck,cch) # can take time for k>2 & h>2 

# chart the di-cusum 

cmu<-mean(y[sel.refy],na.rm=T); csd<-sd(y[sel.refy],na.rm=T)
phi<-cusum.decide(tab$Year,y,cmu,csd,cck,cch)

# convert phi+/phi- of dicusum into a time series of alarm signal 
alarm<-rep(0,length(phi[,1]))
alarm[phi[,4]< -cch]<-phi[phi[,4]< -cch,4]
alarm[phi[,3]>  cch]<-phi[phi[,3]>  cch,3]

# save results in object resnam.xx where xx=jcol    

res<-list("indic"=colnam[jcol],"ref.years"=tab$Year[sel.refy],"ms"=c("m"=cmu,"sd"=csd),
                  "tuning"=param,"phi"=cbind(tab$Year,alarm))
assign(paste(resnam,jcol,sep="."),res)


# ======== once all indicators have been cusumed                     
# ======== create summary tables of alarm series and cusum parameters for all indicators 
# ======== save summary tables to disk

# numcol: index of columns in the tab2 for the indicators 
# modify numcol as necessary 
numcol<-5:11

# create a diagnostic table : columns are the numcol indicator alarm series
# name of table = traflit. modify as necessary 
traflit<-matrix(NA,nrow=length(tab$Year),ncol=(length(numcol)+1))
dimnames(traflit)<-list(NULL,c("Year",colnam[numcol]))
traflit[,1]<-tab$Year
j2<-2
for (j in 1:length(numcol)) {
  j1<-numcol[j]
  traflit[,j2]<-get(paste(resnam,j1,sep="."))$phi[,2] 
  j2<-j2+1 }
write(traflit,file="cusumtraflit",sep="\t",row.names=F)

# create a table of cusum parameters : lines are for the numcol indicators
#                                      10 parameters
# name of table = cusumparam. modify as necessary 
tabc<-matrix(NA,nrow=length(numcol),ncol=2)
tabn<-matrix(NA,nrow=length(numcol),ncol=8)
for (j in 1:length(numcol)) {
  j1<-numcol[j] ; x<-get(paste(resnam,j1,sep="."))
  tabc[j,]<-c(x[[1]],toString(x[[2]]))
  tabn[j,]<-c(x[[3]],x[[4]]) 
}
cusumparam<-data.frame(tabc,tabn) 
dimnames(cusumparam)<-list(colnam[numcol],c("indic","ref.years","mu","sd",
                      "k","h","ic.arl","ic.rl.25","shift","oc.arl"))

write.table(cusumparam,file="cusumparam.txt",sep="\t",row.names=F)



